@@ -1,14 +1,16 @@
 from copy import copy, deepcopy
-from datetime import datetime
+import datetime as dt
 
 import pytest
 
 from bson import ObjectId, DBRef
-from marshmallow import validate
-
-from umongo import (Document, EmbeddedDocument, Schema, fields, exceptions,
-                    post_dump, pre_load, validates_schema)
+import marshmallow as ma
 
+from umongo import (
+    Document, EmbeddedDocument, MixinDocument,
+    fields, exceptions, post_dump, pre_load, validates_schema, ExposeMissing
+)
+from umongo.abstract import BaseSchema
 from .common import BaseTest
 
 
@@ -42,17 +44,17 @@ class TestDocument(BaseTest):
 
     def test_repr(self):
         # I love readable stuff !
-        john = self.Student(name='John Doe', birthday=datetime(1995, 12, 12), gpa=3.0)
+        john = self.Student(name='John Doe', birthday=dt.datetime(1995, 12, 12), gpa=3.0)
         assert 'tests.test_document.Student' in repr(john)
         assert 'name' in repr(john)
         assert 'birthday' in repr(john)
         assert 'gpa' in repr(john)
 
     def test_create(self):
-        john = self.Student(name='John Doe', birthday=datetime(1995, 12, 12), gpa=3.0)
+        john = self.Student(name='John Doe', birthday=dt.datetime(1995, 12, 12), gpa=3.0)
         assert john.to_mongo() == {
             'name': 'John Doe',
-            'birthday': datetime(1995, 12, 12),
+            'birthday': dt.datetime(1995, 12, 12),
             'gpa': 3.0
         }
         assert john.is_created is False
@@ -61,37 +63,37 @@ class TestDocument(BaseTest):
 
     def test_from_mongo(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john.to_mongo(update=True) is None
         assert john.is_created is True
         assert john.to_mongo() == {
             'name': 'John Doe',
-            'birthday': datetime(1995, 12, 12),
+            'birthday': dt.datetime(1995, 12, 12),
             'gpa': 3.0
         }
 
     def test_update(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         john.name = 'William Doe'
-        john.birthday = datetime(1996, 12, 12)
+        john.birthday = dt.datetime(1996, 12, 12)
         assert john.to_mongo(update=True) == {
-            '$set': {'name': 'William Doe', 'birthday': datetime(1996, 12, 12)}}
+            '$set': {'name': 'William Doe', 'birthday': dt.datetime(1996, 12, 12)}}
         john.clear_modified()
         assert john.to_mongo(update=True) is None
 
     def test_dump(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john.dump() == {
             'name': 'John Doe',
-            'birthday': '1995-12-12T00:00:00+00:00',
+            'birthday': '1995-12-12T00:00:00',
             'gpa': 3.0
         }
 
     def test_fields_by_attr(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john.name == 'John Doe'
         john.name = 'William Doe'
         assert john.name == 'William Doe'
@@ -99,16 +101,19 @@ class TestDocument(BaseTest):
         assert john.name is None
         with pytest.raises(AttributeError):
             john.missing
-        with pytest.raises(AttributeError):
-            john.missing = None
         with pytest.raises(AttributeError):
             del john.missing
         with pytest.raises(AttributeError):
-            del john.commit
+            del john.dump
+        john.dummy = 42
+        assert john.dummy == 42
+        del john.dummy
+        with pytest.raises(AttributeError):
+            john.dummy
 
     def test_fields_by_items(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john['name'] == 'John Doe'
         john['name'] = 'William Doe'
         assert john['name'] == 'William Doe'
@@ -146,12 +151,12 @@ class TestDocument(BaseTest):
 
     def test_pk(self):
         john = self.Student.build_from_mongo(data={
-            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john.pk is None
         john_id = ObjectId("5672d47b1d41c88dcd37ef05")
         john = self.Student.build_from_mongo(data={
             '_id': john_id, 'name': 'John Doe',
-            'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
         assert john.pk == john_id
 
         # Don't do that in real life !
@@ -181,12 +186,12 @@ class TestDocument(BaseTest):
 
     def test_equality(self):
         john_data = {
-            '_id': 42, 'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0
+            '_id': 42, 'name': 'John Doe', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0
         }
         john = self.EasyIdStudent.build_from_mongo(data=john_data)
         john2 = self.EasyIdStudent.build_from_mongo(data=john_data)
         phillipe = self.EasyIdStudent.build_from_mongo(data={
-            '_id': 3, 'name': 'Phillipe J. Fry', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
+            '_id': 3, 'name': 'Phillipe J. Fry', 'birthday': dt.datetime(1995, 12, 12), 'gpa': 3.0})
 
         assert john != phillipe
         assert john2 == john
@@ -202,11 +207,11 @@ class TestDocument(BaseTest):
     def test_required_fields(self):
         # Should be able to instanciate document without their required fields
         student = self.Student()
-        with pytest.raises(exceptions.ValidationError):
+        with pytest.raises(ma.ValidationError):
             student.required_validate()
 
         student = self.Student(gpa=2.8)
-        with pytest.raises(exceptions.ValidationError):
+        with pytest.raises(ma.ValidationError):
             student.required_validate()
 
         student = self.Student(gpa=2.8, name='Marty')
@@ -217,14 +222,12 @@ class TestDocument(BaseTest):
 
         @self.instance.register
         class AutoId(Document):
-
-            class Meta:
-                allow_inheritance = True
+            pass
 
         assert 'id' in AutoId.schema.fields
 
         # default id field is only dumpable
-        with pytest.raises(exceptions.ValidationError):
+        with pytest.raises(ma.ValidationError):
             AutoId(id=my_id)
 
         autoid = AutoId.build_from_mongo({'_id': my_id})
@@ -245,14 +248,11 @@ class TestDocument(BaseTest):
         class CustomId(Document):
             int_id = fields.IntField(attribute='_id')
 
-            class Meta:
-                allow_inheritance = True
-
         assert 'id' not in CustomId.schema.fields
-        with pytest.raises(exceptions.ValidationError):
+        with pytest.raises(ma.ValidationError):
             CustomId(id=my_id)
         customid = CustomId(int_id=42)
-        with pytest.raises(exceptions.ValidationError):
+        with pytest.raises(ma.ValidationError):
             customid.int_id = my_id
         assert customid.int_id == 42
         assert customid.pk == customid.int_id
@@ -310,8 +310,7 @@ class TestDocument(BaseTest):
         # a template instead of from an implementation
 
         class ParentAsTemplate(Document):
-            class Meta:
-                allow_inheritance = True
+            pass
 
         Parent = self.instance.register(ParentAsTemplate)
 
@@ -324,18 +323,15 @@ class TestDocument(BaseTest):
     def test_grand_child_inheritance(self):
         @self.instance.register
         class GrandParent(Document):
-            class Meta:
-                allow_inheritance = True
+            pass
 
         @self.instance.register
         class Parent(GrandParent):
-            class Meta:
-                allow_inheritance = True
+            pass
 
         @self.instance.register
         class Uncle(GrandParent):
-            class Meta:
-                allow_inheritance = True
+            pass
 
         @self.instance.register
         class Child(Parent):
@@ -351,7 +347,7 @@ class TestDocument(BaseTest):
         assert Child.opts.offspring == set()
         assert Cousin.opts.offspring == set()
 
-    def test_instanciate_template(self):
+    def test_instantiate_template(self):
 
         class Doc(Document):
             pass
@@ -392,13 +388,13 @@ class TestDocument(BaseTest):
             child = fields.EmbeddedField(Child)
 
         john = Parent(name='John Doe', child={'name': 'John Jr.'})
-        john.birthday = datetime(1995, 12, 12)
+        john.birthday = dt.datetime(1995, 12, 12)
         john.id = ObjectId("5672d47b1d41c88dcd37ef05")
         jane = john.clone()
         assert isinstance(jane, Parent)
         assert isinstance(jane.child, Child)
         assert jane.id is None
-        assert jane.birthday == datetime(1995, 12, 12)
+        assert jane.birthday == dt.datetime(1995, 12, 12)
         assert jane.name == 'John Doe'
         assert jane.child == john.child
         assert jane.child is not john.child
@@ -419,15 +415,149 @@ class TestDocument(BaseTest):
         assert jane.id != john.id
         assert jane.name == 'John Doe'
 
-    def test_validate_default(self):
-        """Check default values are validated"""
+    def test_modify_pk_field(self):
 
-        with pytest.raises(exceptions.ValidationError):
-            class User(Document):
-                name = fields.StringField(
-                    default='Eric',
-                    validate=validate.OneOf(('Stan', 'Kyle', 'Kenny'))
-                )
+        @self.instance.register
+        class User(Document):
+            primary_key = fields.ObjectIdField(attribute='_id', default=ObjectId)
+            name = fields.StrField()
+
+        john = User()
+        john.primary_key = ObjectId()
+        john.from_mongo({'name': 'John Doc'})
+        assert john.is_created
+        with pytest.raises(exceptions.AlreadyCreatedError):
+            john.primary_key = ObjectId()
+        with pytest.raises(exceptions.AlreadyCreatedError):
+            john['primary_key'] = ObjectId()
+        with pytest.raises(exceptions.AlreadyCreatedError):
+            del john.primary_key
+        with pytest.raises(exceptions.AlreadyCreatedError):
+            del john['primary_key']
+        with pytest.raises(exceptions.AlreadyCreatedError):
+            john.update({'primary_key': ObjectId()})
+
+    def test_expose_missing(self):
+        john = self.Student(name='John Doe')
+        assert john.name == 'John Doe'
+        assert john.birthday is None
+        with ExposeMissing():
+            assert john.name == 'John Doe'
+            assert john.birthday is ma.missing
+
+    def test_mixin(self):
+
+        @self.instance.register
+        class PMixin(MixinDocument):
+            pm = fields.IntField()
+
+        @self.instance.register
+        class CMixin(MixinDocument):
+            cm = fields.IntField()
+
+        @self.instance.register
+        class Parent(Document, PMixin):
+            p = fields.StrField()
+
+            class Meta:
+                allow_inheritance = True
+
+        @self.instance.register
+        class Child(Parent, CMixin):
+            c = fields.StrField()
+
+        assert set(Parent.schema.fields.keys()) == {'id', 'p', 'pm'}
+        assert set(Child.schema.fields.keys()) == {'id', 'cls', 'p', 'pm', 'c', 'cm'}
+
+        parent_data = {'p': 'parent', 'pm': 42}
+        child_data = {'c': 'child', 'cm': 12, **parent_data}
+
+        assert Parent(**parent_data).dump() == parent_data
+        assert Child(**child_data).dump() == {**child_data, 'cls': 'Child'}
+
+        parent = Parent()
+        parent.p = 'parent'
+        parent.pm = 42
+        assert parent.p == 'parent'
+        assert parent.pm == 42
+        assert parent.dump() == parent_data
+        del parent.p
+        del parent.pm
+        assert parent.p is None
+        assert parent.pm is None
+
+        parent = Parent()
+        parent['p'] = 'parent'
+        parent['pm'] = 42
+        assert parent['p'] == 'parent'
+        assert parent['pm'] == 42
+        assert parent.dump() == parent_data
+        del parent['p']
+        del parent['pm']
+        assert parent['p'] is None
+        assert parent['pm'] is None
+
+        child = Child()
+        child.c = 'child'
+        child.cm = 12
+        child.p = 'parent'
+        child.pm = 42
+        assert child.c == 'child'
+        assert child.cm == 12
+        assert child.dump() == {'cls': 'Child', **child_data}
+        del child.c
+        del child.cm
+        assert child.c is None
+        assert child.cm is None
+
+        child = Child()
+        child['c'] = 'child'
+        child['cm'] = 12
+        child['p'] = 'parent'
+        child['pm'] = 42
+        assert child['c'] == 'child'
+        assert child['cm'] == 12
+        assert child.dump() == {'cls': 'Child', **child_data}
+        del child['c']
+        del child['cm']
+        assert child['c'] is None
+        assert child['cm'] is None
+
+    def test_mixin_override(self):
+
+        @self.instance.register
+        class PMixin(MixinDocument):
+            pm = fields.IntField()
+
+        @self.instance.register
+        class CMixin(MixinDocument):
+            cm = fields.IntField()
+
+        @self.instance.register
+        class Parent(Document, PMixin):
+            p = fields.StrField()
+            pm = fields.IntField(validate=ma.validate.Range(0, 5))
+
+            class Meta:
+                allow_inheritance = True
+
+        @self.instance.register
+        class Child(Parent, CMixin):
+            c = fields.StrField()
+            cm = fields.IntField(validate=ma.validate.Range(0, 5))
+
+        assert set(Parent.schema.fields.keys()) == {'id', 'p', 'pm'}
+        assert set(Child.schema.fields.keys()) == {'id', 'cls', 'p', 'pm', 'c', 'cm'}
+
+        parent_data = {'p': 'parent', 'pm': 42}
+        child_data = {'c': 'Child', 'cm': 12, **parent_data}
+
+        with pytest.raises(ma.ValidationError) as exc:
+            Parent(**parent_data)
+        assert set(exc.value.messages.keys()) == {'pm'}
+        with pytest.raises(ma.ValidationError) as exc:
+            Child(**child_data)
+        assert set(exc.value.messages.keys()) == {'pm', 'cm'}
 
 
 class TestConfig(BaseTest):
@@ -440,7 +570,7 @@ class TestConfig(BaseTest):
             pass
 
         d = Doc()
-        assert isinstance(d.schema, Schema)
+        assert isinstance(d.schema, BaseSchema)
 
     def test_base_config(self):
 
@@ -450,7 +580,6 @@ class TestConfig(BaseTest):
 
         assert Doc.opts.collection_name == 'doc'
         assert Doc.opts.abstract is False
-        assert Doc.opts.allow_inheritance is False
         assert Doc.opts.instance is self.instance
         assert Doc.opts.is_child is False
         assert Doc.opts.indexes == []
@@ -468,7 +597,6 @@ class TestConfig(BaseTest):
         class DocChild1(AbsDoc):
 
             class Meta:
-                allow_inheritance = True
                 collection_name = 'col1'
 
         @self.instance.register
@@ -481,11 +609,22 @@ class TestConfig(BaseTest):
             class Meta:
                 collection_name = 'col2'
 
-        assert DocChild1.opts.collection_name is 'col1'
-        assert DocChild1Child.opts.collection_name is 'col1'
-        assert DocChild1Child.opts.allow_inheritance is False
+        assert DocChild1.opts.collection_name == 'col1'
+        assert DocChild1Child.opts.collection_name == 'col1'
         assert DocChild2.opts.collection_name == 'col2'
 
+    def test_inheritance_from_embedded_document(self):
+
+        @self.instance.register
+        class Parent(EmbeddedDocument):
+            last_name = fields.StrField()
+
+        @self.instance.register
+        class Child(Parent):
+            first_name = fields.StrField()
+
+        Child(first_name='John', last_name='Doe')
+
     def test_marshmallow_tags_build(self):
 
         @self.instance.register
@@ -493,7 +632,7 @@ class TestConfig(BaseTest):
             name = fields.StringField()
 
             @pre_load
-            def test(self, data):
+            def test(self, data, **kwargs):
                 return data
 
         Animal(name='Scruffy')
@@ -504,9 +643,6 @@ class TestConfig(BaseTest):
         class Animal(Document):
             name = fields.StrField(attribute='_id')  # Overwrite automatic pk
 
-            class Meta:
-                allow_inheritance = True
-
         @self.instance.register
         class Dog(Animal):
             pass
@@ -514,19 +650,19 @@ class TestConfig(BaseTest):
         @self.instance.register
         class Duck(Animal):
             @post_dump
-            def dump_custom_cls_name(self, data):
+            def dump_custom_cls_name(self, data, **kwargs):
                 data['race'] = data.pop('cls')
                 return data
 
             @pre_load
-            def load_custom_cls_name(self, data):
+            def load_custom_cls_name(self, data, **kwargs):
                 data.pop('race', None)
                 return data
 
             @validates_schema(pass_original=True)
-            def custom_validate(self, data, original_data):
+            def custom_validate(self, data, original_data, **kwargs):
                 if original_data['name'] != 'Donald':
-                    raise exceptions.ValidationError('Not suitable name for duck !', 'name')
+                    raise ma.ValidationError('Not suitable name for duck !', 'name')
 
         duck = Duck(name='Donald')
         dog = Dog(name='Pluto')
@@ -536,55 +672,30 @@ class TestConfig(BaseTest):
         assert dog.dump() == {'name': 'Pluto', 'cls': 'Dog'}
         assert Duck(name='Donald', race='Duck')._data == duck._data
 
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             Duck(name='Roger')
         exc.value.args[0] == {'name': 'Not suitable name for duck !'}
 
     def test_bad_inheritance(self):
-        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
-            @self.instance.register
-            class BadAbstractDoc(Document):
-                class Meta:
-                    allow_inheritance = False
-                    abstract = True
-        assert exc.value.args[0] == "Abstract document cannot disable inheritance"
-
-        @self.instance.register
-        class NotParent(Document):
-            pass
-
-        assert not NotParent.opts.allow_inheritance
-
-        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
-            @self.instance.register
-            class ImpossibleChildDoc1(NotParent):
-                pass
-        assert exc.value.args[0] == ("Document"
-            " <Implementation class 'tests.test_document.NotParent'>"
-            " doesn't allow inheritance")
-
         @self.instance.register
         class NotAbstractParent(Document):
-            class Meta:
-                allow_inheritance = True
+            pass
 
         with pytest.raises(exceptions.DocumentDefinitionError) as exc:
             @self.instance.register
             class ImpossibleChildDoc2(NotAbstractParent):
                 class Meta:
                     abstract = True
-        assert exc.value.args[0] == "Abstract document should have all it parents abstract"
+        assert exc.value.args[0] == "Abstract document should have all its parents abstract"
 
         @self.instance.register
         class ParentWithCol1(Document):
             class Meta:
-                allow_inheritance = True
                 collection_name = 'col1'
 
         @self.instance.register
         class ParentWithCol2(Document):
             class Meta:
-                allow_inheritance = True
                 collection_name = 'col2'
 
         with pytest.raises(exceptions.DocumentDefinitionError) as exc:
@@ -622,6 +733,6 @@ class TestConfig(BaseTest):
         assert non_strict_doc.to_mongo() == data_with_bonus
         non_strict_doc.dump() == {'a': 42}
 
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             NonStrictDoc(a=42, b='foo')
-        assert exc.value.messages == {'_schema': ['Unknown field name b.']}
+        assert exc.value.messages == {'b': ['Unknown field.']}
