@@ -2,10 +2,10 @@ from copy import copy, deepcopy
 
 import pytest
 
-from marshmallow import ValidationError, missing
+import marshmallow as ma
 
 from umongo.data_proxy import data_proxy_factory
-from umongo import Document, EmbeddedDocument, fields, exceptions
+from umongo import Document, EmbeddedDocument, MixinDocument, fields, exceptions, ExposeMissing
 
 from .common import BaseTest
 
@@ -30,6 +30,9 @@ class TestEmbeddedDocument(BaseTest):
         })
 
         assert document is not None
+        assert isinstance(document, MyDoc)
+        assert isinstance(document.embedded, MyParentEmbeddedDocument)
+        assert isinstance(document.embedded.embedded, MyChildEmbeddedDocument)
 
     def test_embedded_document(self):
         @self.instance.register
@@ -100,7 +103,7 @@ class TestEmbeddedDocument(BaseTest):
         assert embedded_doc.a == 1
         assert embedded_doc.b == 42
 
-        with pytest.raises(ValidationError):
+        with pytest.raises(ma.ValidationError):
             MyEmbeddedDocument(in_mongo_a=1, b=2)
 
         embedded_doc['a'] = 1
@@ -115,26 +118,57 @@ class TestEmbeddedDocument(BaseTest):
         assert "'a': 1" in repr_d
         assert "'b': 2" in repr_d
 
-        # Test unknown fields
-        with pytest.raises(AttributeError):
-            embedded_doc.dummy
-        with pytest.raises(AttributeError):
-            embedded_doc.dummy = None
-        with pytest.raises(AttributeError):
-            del embedded_doc.dummy
-        with pytest.raises(KeyError):
-            embedded_doc['dummy']
-        with pytest.raises(KeyError):
-            embedded_doc['dummy'] = None
-        with pytest.raises(KeyError):
-            del embedded_doc['dummy']
-
         # Test allow_none
         d3 = MyDataProxy({'embedded': None})
         assert d3.to_mongo() == {'in_mongo_embedded': None}
         d3.from_mongo({'in_mongo_embedded': None})
         assert d3.get('embedded') is None
 
+    def test_fields_by_attr(self):
+        @self.instance.register
+        class EmbeddedStudent(EmbeddedDocument):
+            name = fields.StrField(required=True)
+            birthday = fields.DateTimeField()
+            gpa = fields.FloatField()
+
+        john = EmbeddedStudent.build_from_mongo(data={'name': 'John Doe'})
+        assert john.name == 'John Doe'
+        john.name = 'William Doe'
+        assert john.name == 'William Doe'
+        del john.name
+        assert john.name is None
+        with pytest.raises(AttributeError):
+            john.missing
+        with pytest.raises(AttributeError):
+            del john.missing
+        with pytest.raises(AttributeError):
+            del john.dump
+        john.dummy = 42
+        assert john.dummy == 42
+        del john.dummy
+        with pytest.raises(AttributeError):
+            john.dummy
+
+    def test_fields_by_items(self):
+        @self.instance.register
+        class EmbeddedStudent(EmbeddedDocument):
+            name = fields.StrField(required=True)
+            birthday = fields.DateTimeField()
+            gpa = fields.FloatField()
+
+        john = EmbeddedStudent.build_from_mongo(data={'name': 'John Doe'})
+        assert john['name'] == 'John Doe'
+        john['name'] = 'William Doe'
+        assert john['name'] == 'William Doe'
+        del john['name']
+        assert john['name'] is None
+        with pytest.raises(KeyError):
+            john['missing']
+        with pytest.raises(KeyError):
+            john['missing'] = None
+        with pytest.raises(KeyError):
+            del john['missing']
+
     def test_bad_embedded_document(self):
 
         @self.instance.register
@@ -144,34 +178,34 @@ class TestEmbeddedDocument(BaseTest):
         @self.instance.register
         class MyDoc(Document):
             e = fields.EmbeddedField(MyEmbeddedDocument)
-            l = fields.ListField(fields.EmbeddedField(MyEmbeddedDocument))
+            li = fields.ListField(fields.EmbeddedField(MyEmbeddedDocument))
             b = fields.IntField(required=True)
 
-        with pytest.raises(ValidationError) as exc:
-            MyDoc(l={})
-        assert exc.value.args[0] == {'l': ['Not a valid list.']}
+        with pytest.raises(ma.ValidationError) as exc:
+            MyDoc(li={})
+        assert exc.value.args[0] == {'li': ['Not a valid list.']}
 
-        with pytest.raises(ValidationError) as exc:
-            MyDoc(l=True)
-        assert exc.value.args[0] == {'l': ['Not a valid list.']}
+        with pytest.raises(ma.ValidationError) as exc:
+            MyDoc(li=True)
+        assert exc.value.args[0] == {'li': ['Not a valid list.']}
 
-        with pytest.raises(ValidationError) as exc:
-            MyDoc(l="string is not a list")
-        assert exc.value.args[0] == {'l': ['Not a valid list.']}
+        with pytest.raises(ma.ValidationError) as exc:
+            MyDoc(li="string is not a list")
+        assert exc.value.args[0] == {'li': ['Not a valid list.']}
 
-        with pytest.raises(ValidationError) as exc:
-            MyDoc(l=[42])
-        assert exc.value.args[0] == {'l': {0: {'_schema': ['Invalid input type.']}}}
+        with pytest.raises(ma.ValidationError) as exc:
+            MyDoc(li=[42])
+        assert exc.value.args[0] == {'li': {0: {'_schema': ['Invalid input type.']}}}
 
-        with pytest.raises(ValidationError) as exc:
-            MyDoc(l=[{}, 42])
-        assert exc.value.args[0] == {'l': {1: {'_schema': ['Invalid input type.']}}}
+        with pytest.raises(ma.ValidationError) as exc:
+            MyDoc(li=[{}, 42])
+        assert exc.value.args[0] == {'li': {1: {'_schema': ['Invalid input type.']}}}
 
-        with pytest.raises(ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             MyDoc(b=[{}])
         assert exc.value.args[0] == {'b': ['Not a valid integer.']}
 
-        with pytest.raises(ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             MyDoc(e=[{}])
         assert exc.value.args[0] == {'e': {'_schema': ['Invalid input type.']}}
 
@@ -212,10 +246,10 @@ class TestEmbeddedDocument(BaseTest):
         assert child.to_mongo() == {'in_mongo_a_child': 1, 'b': 2, 'c': 3, '_cls': 'EmbeddedChild'}
         assert grandchild.to_mongo() == {'d': 4, '_cls': 'GrandChild'}
 
-        with pytest.raises(ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             MyDoc(parent=OtherEmbedded())
         assert exc.value.args[0] == {'parent': {'_schema': ['Invalid input type.']}}
-        with pytest.raises(ValidationError):
+        with pytest.raises(ma.ValidationError):
             MyDoc(child=parent)
         doc = MyDoc(parent=child, child=child)
         assert doc.child == doc.parent
@@ -225,12 +259,12 @@ class TestEmbeddedDocument(BaseTest):
         assert doc.child.to_mongo() == {'in_mongo_a_child': 1, '_cls': 'GrandChild'}
         assert doc.parent.to_mongo() == {'in_mongo_a_child': 1, '_cls': 'EmbeddedChild'}
 
-        with pytest.raises(ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             MyDoc(child={'a': 1, '_cls': 'GrandChild'})
-        assert exc.value.messages == {'child': {'_schema': ['Unknown field name _cls.']}}
+        assert exc.value.messages == {'child': {'_cls': ['Unknown field.']}}
 
         # Try to build a non-child document
-        with pytest.raises(ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             MyDoc(child={'cls': 'OtherEmbedded'})
         assert exc.value.messages == {'child': ['Unknown document `OtherEmbedded`.']}
 
@@ -270,9 +304,6 @@ class TestEmbeddedDocument(BaseTest):
         class ConcreteChild(AbstractParent, AlienClass):
             c = fields.IntField()
 
-            class Meta:
-                allow_inheritance = True
-
         @self.instance.register
         class ConcreteGrandChild(AbstractChild):
             d = fields.IntField()
@@ -294,46 +325,41 @@ class TestEmbeddedDocument(BaseTest):
         cgc = ConcreteGrandChild(a=1, b=2, c=3, d=4)
         ccgc = ConcreteConcreteGrandChild(a=1, b=2, c=3, d=4)
 
-        assert cc.to_mongo() == {'in_mongo_a_parent': 1, 'b': 2, 'c': 3, '_cls': 'ConcreteChild'}
-        assert cgc.to_mongo() == {'in_mongo_a_child': 1, 'b': 2, 'c': 3, 'd': 4, '_cls': 'ConcreteGrandChild'}
-        assert ccgc.to_mongo() == {'in_mongo_a_parent': 1, 'b': 2, 'c': 3, 'd': 4, '_cls': 'ConcreteConcreteGrandChild'}
+        # Child of abstract doesn't need `cls` hint field in serialization
+        assert cc.to_mongo() == {'in_mongo_a_parent': 1, 'b': 2, 'c': 3}
+        assert cgc.to_mongo() == {'in_mongo_a_child': 1, 'b': 2, 'c': 3, 'd': 4}
+        # But child of non abstract does
+        assert ccgc.to_mongo() == {
+            'in_mongo_a_parent': 1, 'b': 2, 'c': 3, 'd': 4, '_cls': 'ConcreteConcreteGrandChild'}
+
+        # Cannot use abstract embedded document in EmbeddedField
+        @self.instance.register
+        class MyDoc(Document):
+            impossible = fields.EmbeddedField(AbstractParent)
+
+        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
+            MyDoc(impossible={"a": 12})
+        assert exc.value.args[0] == "EmbeddedField doesn't accept abstract embedded document"
+
+        @self.instance.register
+        class MyOtherDoc(Document):
+            impossible = fields.EmbeddedField(AbstractChild)
+
+        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
+            MyOtherDoc(impossible={"a": 12})
+        assert exc.value.args[0] == "EmbeddedField doesn't accept abstract embedded document"
 
     def test_bad_inheritance(self):
-        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
-            @self.instance.register
-            class BadAbstract(EmbeddedDocument):
-                class Meta:
-                    allow_inheritance = False
-                    abstract = True
-        assert exc.value.args[0] == "Abstract embedded document cannot disable inheritance"
-
-        @self.instance.register
-        class NotParent(EmbeddedDocument):
-            class Meta:
-                allow_inheritance = False
-
-        with pytest.raises(exceptions.DocumentDefinitionError) as exc:
-            @self.instance.register
-            class ImpossibleChild1(NotParent):
-                pass
-        assert exc.value.args[0] == ("EmbeddedDocument"
-            " <Implementation class 'tests.test_embedded_document.NotParent'>"
-            " doesn't allow inheritance")
-
         @self.instance.register
         class NotAbstractParent(EmbeddedDocument):
-            class Meta:
-                allow_inheritance = True
-
-        # Unlike Document, EmbeddedDocument should allow inheritance by default
-        assert NotAbstractParent.opts.allow_inheritance
+            pass
 
         with pytest.raises(exceptions.DocumentDefinitionError) as exc:
             @self.instance.register
             class ImpossibleChild2(NotAbstractParent):
                 class Meta:
                     abstract = True
-        assert exc.value.args[0] == "Abstract embedded document should have all it parents abstract"
+        assert exc.value.args[0] == "Abstract document should have all its parents abstract"
 
     def test_property(self):
         @self.instance.register
@@ -361,24 +387,28 @@ class TestEmbeddedDocument(BaseTest):
     def test_equality(self):
         @self.instance.register
         class MyChildEmbeddedDocument(EmbeddedDocument):
-            num = fields.IntField()
+            num_1 = fields.IntField()
+            num_2 = fields.IntField()
 
         @self.instance.register
         class MyParentEmbeddedDocument(EmbeddedDocument):
             embedded = fields.EmbeddedField(MyChildEmbeddedDocument)
 
-        emb_1 = MyParentEmbeddedDocument(embedded={'num': 1})
-        emb_2 = MyParentEmbeddedDocument(embedded={'num': 1})
+        emb_1 = MyParentEmbeddedDocument(embedded={'num_1': 1, 'num_2': 2})
+        emb_2 = MyParentEmbeddedDocument(embedded={'num_1': 1, 'num_2': 2})
         emb_3 = MyParentEmbeddedDocument(embedded={})
         emb_4 = MyParentEmbeddedDocument()
+        emb_5 = MyParentEmbeddedDocument(embedded={'num_2': 2})
+        emb_5["embedded"]["num_1"] = 1
 
         assert emb_1 == emb_2
         assert emb_1 != emb_3
         assert emb_1 != emb_4
+        assert emb_1 == emb_5
         assert emb_1 != None  # noqa: E711 (None comparison)
-        assert emb_1 != missing
+        assert emb_1 != ma.missing
         assert None != emb_1  # noqa: E711 (None comparison)
-        assert missing != emb_1
+        assert ma.missing != emb_1
 
     def test_strict_embedded_document(self):
         @self.instance.register
@@ -400,9 +430,9 @@ class TestEmbeddedDocument(BaseTest):
         assert non_strict_doc.to_mongo() == data_with_bonus
         non_strict_doc.dump() == {'a': 42}
 
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             NonStrictEmbeddedDoc(a=42, b='foo')
-        assert exc.value.messages == {'_schema': ['Unknown field name b.']}
+        assert exc.value.messages == {'b': ['Unknown field.']}
 
     def test_deepcopy(self):
 
@@ -423,3 +453,137 @@ class TestEmbeddedDocument(BaseTest):
         assert jane.name == john.name
         assert jane.child == john.child
         assert jane.child is not john.child
+
+    def test_expose_missing(self):
+        @self.instance.register
+        class Child(EmbeddedDocument):
+            name = fields.StrField()
+            age = fields.IntField()
+
+        @self.instance.register
+        class Parent(EmbeddedDocument):
+            name = fields.StrField()
+            child = fields.EmbeddedField(Child)
+
+        parent = Parent(**{'child': {'age': 42}})
+        assert parent.name is None
+        assert parent.child.name is None
+        assert parent.child.age == 42
+        with ExposeMissing():
+            assert parent.name is ma.missing
+            assert parent.child.name is ma.missing
+            assert parent.child.age == 42
+
+    def test_mixin(self):
+
+        @self.instance.register
+        class PMixin(MixinDocument):
+            pm = fields.IntField()
+
+        @self.instance.register
+        class CMixin(MixinDocument):
+            cm = fields.IntField()
+
+        @self.instance.register
+        class Parent(EmbeddedDocument, PMixin):
+            p = fields.StrField()
+
+            class Meta:
+                allow_inheritance = True
+
+        @self.instance.register
+        class Child(Parent, CMixin):
+            c = fields.StrField()
+
+        assert set(Parent.schema.fields.keys()) == {'p', 'pm'}
+        assert set(Child.schema.fields.keys()) == {'cls', 'p', 'pm', 'c', 'cm'}
+
+        parent_data = {'p': 'parent', 'pm': 42}
+        child_data = {'c': 'child', 'cm': 12, **parent_data}
+
+        assert Parent(**parent_data).dump() == parent_data
+        assert Child(**child_data).dump() == {**child_data, 'cls': 'Child'}
+
+        parent = Parent()
+        parent.p = 'parent'
+        parent.pm = 42
+        assert parent.p == 'parent'
+        assert parent.pm == 42
+        assert parent.dump() == parent_data
+        del parent.p
+        del parent.pm
+        assert parent.p is None
+        assert parent.pm is None
+
+        parent = Parent()
+        parent['p'] = 'parent'
+        parent['pm'] = 42
+        assert parent['p'] == 'parent'
+        assert parent['pm'] == 42
+        assert parent.dump() == parent_data
+        del parent['p']
+        del parent['pm']
+        assert parent['p'] is None
+        assert parent['pm'] is None
+
+        child = Child()
+        child.c = 'child'
+        child.cm = 12
+        child.p = 'parent'
+        child.pm = 42
+        assert child.c == 'child'
+        assert child.cm == 12
+        assert child.dump() == {'cls': 'Child', **child_data}
+        del child.c
+        del child.cm
+        assert child.c is None
+        assert child.cm is None
+
+        child = Child()
+        child['c'] = 'child'
+        child['cm'] = 12
+        child['p'] = 'parent'
+        child['pm'] = 42
+        assert child['c'] == 'child'
+        assert child['cm'] == 12
+        assert child.dump() == {'cls': 'Child', **child_data}
+        del child['c']
+        del child['cm']
+        assert child['c'] is None
+        assert child['cm'] is None
+
+    def test_mixin_override(self):
+
+        @self.instance.register
+        class PMixin(MixinDocument):
+            pm = fields.IntField()
+
+        @self.instance.register
+        class CMixin(MixinDocument):
+            cm = fields.IntField()
+
+        @self.instance.register
+        class Parent(EmbeddedDocument, PMixin):
+            p = fields.StrField()
+            pm = fields.IntField(validate=ma.validate.Range(0, 5))
+
+            class Meta:
+                allow_inheritance = True
+
+        @self.instance.register
+        class Child(Parent, CMixin):
+            c = fields.StrField()
+            cm = fields.IntField(validate=ma.validate.Range(0, 5))
+
+        assert set(Parent.schema.fields.keys()) == {'p', 'pm'}
+        assert set(Child.schema.fields.keys()) == {'cls', 'p', 'pm', 'c', 'cm'}
+
+        parent_data = {'p': 'parent', 'pm': 42}
+        child_data = {'c': 'Child', 'cm': 12, **parent_data}
+
+        with pytest.raises(ma.ValidationError) as exc:
+            Parent(**parent_data)
+        assert set(exc.value.messages.keys()) == {'pm'}
+        with pytest.raises(ma.ValidationError) as exc:
+            Child(**child_data)
+        assert set(exc.value.messages.keys()) == {'pm', 'cm'}
