@@ -9,15 +9,15 @@ __all__ = ('List', 'Dict', 'Reference')
 
 class List(BaseDataObject, list):
 
-    __slots__ = ('container_field', '_modified')
+    __slots__ = ('inner_field', '_modified')
 
-    def __init__(self, container_field, *args, **kwargs):
+    def __init__(self, inner_field, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self._modified = False
-        self.container_field = container_field
+        self.inner_field = inner_field
 
     def __setitem__(self, key, obj):
-        obj = self.container_field.deserialize(obj)
+        obj = self.inner_field.deserialize(obj)
         super().__setitem__(key, obj)
         self.set_modified()
 
@@ -26,11 +26,17 @@ class List(BaseDataObject, list):
         self.set_modified()
 
     def append(self, obj):
-        obj = self.container_field.deserialize(obj)
+        obj = self.inner_field.deserialize(obj)
         ret = super().append(obj)
         self.set_modified()
         return ret
 
+    def insert(self, i, obj):
+        obj = self.inner_field.deserialize(obj)
+        ret = super().insert(i, obj)
+        self.set_modified()
+        return ret
+
     def pop(self, *args, **kwargs):
         ret = super().pop(*args, **kwargs)
         self.set_modified()
@@ -57,7 +63,7 @@ class List(BaseDataObject, list):
         return ret
 
     def extend(self, iterable):
-        iterable = [self.container_field.deserialize(obj) for obj in iterable]
+        iterable = [self.inner_field.deserialize(obj) for obj in iterable]
         ret = super().extend(iterable)
         self.set_modified()
         return ret
@@ -66,42 +72,88 @@ class List(BaseDataObject, list):
         return '<object %s.%s(%s)>' % (
             self.__module__, self.__class__.__name__, list(self))
 
-    def set_modified(self):
-        self._modified = True
-
     def is_modified(self):
         if self._modified:
             return True
-        if len(self) and isinstance(self[0], BaseDataObject):
-            # Recursive handling needed
+        if self and isinstance(self[0], BaseDataObject):
             return any(obj.is_modified() for obj in self)
         return False
 
+    def set_modified(self):
+        self._modified = True
+
     def clear_modified(self):
         self._modified = False
-        if len(self) and isinstance(self[0], BaseDataObject):
-            # Recursive handling needed
+        if self and isinstance(self[0], BaseDataObject):
             for obj in self:
                 obj.clear_modified()
 
 
-# TODO: Dict is to much raw: you need to use `set_modified` by hand !
 class Dict(BaseDataObject, dict):
 
-    __slots__ = ('_modified', )
+    __slots__ = ('key_field', 'value_field', '_modified')
 
-    def __init__(self, *args, **kwargs):
-        self._modified = False
+    def __init__(self, key_field, value_field, *args, **kwargs):
         super().__init__(*args, **kwargs)
+        self._modified = False
+        self.key_field = key_field
+        self.value_field = value_field
+
+    def __setitem__(self, key, obj):
+        key = self.key_field.deserialize(key) if self.key_field else key
+        obj = self.value_field.deserialize(obj) if self.value_field else obj
+        super().__setitem__(key, obj)
+        self.set_modified()
+
+    def __delitem__(self, key):
+        super().__delitem__(key)
+        self.set_modified()
+
+    def pop(self, *args, **kwargs):
+        ret = super().pop(*args, **kwargs)
+        self.set_modified()
+        return ret
+
+    def popitem(self, *args, **kwargs):
+        ret = super().popitem(*args, **kwargs)
+        self.set_modified()
+        return ret
+
+    def setdefault(self, key, obj=None):
+        key = self.key_field.deserialize(key) if self.key_field else key
+        obj = self.value_field.deserialize(obj) if self.value_field else obj
+        ret = super().setdefault(key, obj)
+        self.set_modified()
+        return ret
+
+    def update(self, other):
+        new = {
+            self.key_field.deserialize(k) if self.key_field else k:
+            self.value_field.deserialize(v) if self.value_field else v
+            for k, v in other.items()
+        }
+        super().update(new)
+        self.set_modified()
+
+    def __repr__(self):
+        return '<object %s.%s(%s)>' % (
+            self.__module__, self.__class__.__name__, dict(self))
 
     def is_modified(self):
-        return self._modified
+        if self._modified:
+            return True
+        if self and any(isinstance(v, BaseDataObject) for v in self.values()):
+            return any(obj.is_modified() for obj in self.values())
+        return False
 
     def set_modified(self):
         self._modified = True
 
     def clear_modified(self):
         self._modified = False
+        if self and any(isinstance(v, BaseDataObject) for v in self.values()):
+            for obj in self.values():
+                obj.clear_modified()
 
 
 class Reference:
@@ -133,8 +185,8 @@ class Reference:
     def __eq__(self, other):
         if isinstance(other, self.document_cls):
             return other.pk == self.pk
-        elif isinstance(other, Reference):
+        if isinstance(other, Reference):
             return self.pk == other.pk and self.document_cls == other.document_cls
-        elif isinstance(other, DBRef):
+        if isinstance(other, DBRef):
             return self.pk == other.id and self.document_cls.collection.name == other.collection
         return NotImplemented
