@@ -1,13 +1,19 @@
+"""umongo Document"""
 from copy import deepcopy
 
 from bson import DBRef
-from marshmallow import pre_load, post_load, pre_dump, post_dump, validates_schema  # republishing
+import marshmallow as ma
+from marshmallow import (
+    pre_load, post_load, pre_dump, post_dump, validates_schema,  # republishing
+)
 
-from .abstract import BaseDataObject
-from .data_proxy import missing
-from .exceptions import (NotCreatedError, NoDBDefinedError,
-                         AbstractDocumentError, DocumentDefinitionError)
-from .template import Implementation, Template, MetaImplementation
+from .exceptions import (
+    AlreadyCreatedError, NotCreatedError, NoDBDefinedError, AbstractDocumentError
+)
+from .template import Template, MetaImplementation
+from .embedded_document import EmbeddedDocumentImplementation
+from .data_objects import Reference
+from .indexes import parse_index
 
 
 __all__ = (
@@ -37,7 +43,6 @@ class DocumentTemplate(Template):
         or `marshmallow.post_dump`) to this class that will be passed
         to the marshmallow schema internally used for this document.
     """
-    pass
 
 
 Document = DocumentTemplate
@@ -67,7 +72,6 @@ class DocumentOpts:
     instance             no                     Implementation's instance
     abstract             yes                    Document has no collection
                                                 and can only be inherited
-    allow_inheritance    yes                    Allow the document to be subclassed
     collection_name      yes                    Name of the collection to store
                                                 the document into
     is_child             no                     Document inherit of a non-abstract document
@@ -76,15 +80,12 @@ class DocumentOpts:
     indexes              yes                    List of custom indexes
     offspring            no                     List of Documents inheriting this one
     ==================== ====================== ===========
-
     """
-
     def __repr__(self):
         return ('<{ClassName}('
                 'instance={self.instance}, '
                 'template={self.template}, '
                 'abstract={self.abstract}, '
-                'allow_inheritance={self.allow_inheritance}, '
                 'collection_name={self.collection_name}, '
                 'is_child={self.is_child}, '
                 'strict={self.strict}, '
@@ -93,23 +94,22 @@ class DocumentOpts:
                 .format(ClassName=self.__class__.__name__, self=self))
 
     def __init__(self, instance, template, collection_name=None, abstract=False,
-                 allow_inheritance=None, indexes=None, is_child=True, strict=True,
-                 offspring=None):
+                 indexes=None, is_child=True, strict=True, offspring=None):
         self.instance = instance
         self.template = template
         self.collection_name = collection_name if not abstract else None
         self.abstract = abstract
-        self.allow_inheritance = abstract if allow_inheritance is None else allow_inheritance
         self.indexes = indexes or []
         self.is_child = is_child
         self.strict = strict
         self.offspring = set(offspring) if offspring else set()
-        if self.abstract and not self.allow_inheritance:
-            raise DocumentDefinitionError("Abstract document cannot disable inheritance")
 
 
 class MetaDocumentImplementation(MetaImplementation):
 
+    def __init__(cls, *args, **kwargs):
+        cls._indexes = None
+
     @property
     def collection(cls):
         """
@@ -121,8 +121,63 @@ class MetaDocumentImplementation(MetaImplementation):
             raise NoDBDefinedError('Instance must be initialized first')
         return cls.opts.instance.db[cls.opts.collection_name]
 
+    @property
+    def indexes(cls):
+        """
+        Retrieve all indexes (custom defined in meta class, by inheritances
+        and unique attributes in fields)
+        """
+        if cls._indexes is None:
 
-class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocumentImplementation):
+            idxs = []
+            is_child = cls.opts.is_child
+
+            # First collect parent indexes (including inherited field's unique indexes)
+            for base in cls.mro():
+                if (
+                        base is not cls and
+                        issubclass(base, DocumentImplementation) and
+                        # Skip base framework doc classes
+                        hasattr(base, "schema")
+                ):
+                    idxs += base.indexes
+
+            # Then get our own custom indexes
+            if hasattr(cls, "Meta") and hasattr(cls.Meta, "indexes"):
+                custom_indexes = [
+                    parse_index(x, base_compound_field="_cls" if is_child else None)
+                    for x in cls.Meta.indexes
+                ]
+                idxs += custom_indexes
+
+            # Add _cls to indexes
+            if is_child:
+                idxs.append(parse_index('_cls'))
+
+            # Finally parse our own fields (i.e. not inherited) for unique indexes
+            def parse_field(mongo_path, path, field):
+                if field.unique:
+                    index = {'unique': True, 'key': [mongo_path]}
+                    if not field.required or field.allow_none:
+                        index['sparse'] = True
+                    if is_child:
+                        index['key'].append('_cls')
+                    idxs.append(parse_index(index))
+
+            for name, field in cls.schema.fields.items():
+                parse_field(name or field.attribute, name, field)
+                if hasattr(field, 'map_to_field'):
+                    field.map_to_field(name or field.attribute, name, parse_field)
+
+            cls._indexes = idxs
+
+        return cls._indexes
+
+
+class DocumentImplementation(
+    EmbeddedDocumentImplementation,
+    metaclass=MetaDocumentImplementation
+):
     """
     Represent a document once it has been implemented inside a
     :class:`umongo.instance.BaseInstance`.
@@ -132,30 +187,27 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
     """
 
     __slots__ = ('is_created', '_data')
-    __real_attributes = None
     opts = DocumentOpts(None, DocumentTemplate, abstract=True)
 
     def __init__(self, **kwargs):
-        super().__init__()
         if self.opts.abstract:
             raise AbstractDocumentError("Cannot instantiate an abstract Document")
         self.is_created = False
         "Return True if the document has been commited to database"  # is_created's docstring
-        self._data = self.DataProxy(kwargs)
+        super().__init__(**kwargs)
 
     def __repr__(self):
         return '<object Document %s.%s(%s)>' % (
             self.__module__, self.__class__.__name__, dict(self._data.items()))
 
     def __eq__(self, other):
-        from .data_objects import Reference
         if self.pk is None:
             return self is other
-        elif isinstance(other, self.__class__) and other.pk is not None:
+        if isinstance(other, self.__class__) and other.pk is not None:
             return self.pk == other.pk
-        elif isinstance(other, DBRef):
+        if isinstance(other, DBRef):
             return other.collection == self.collection.name and other.id == self.pk
-        elif isinstance(other, Reference):
+        if isinstance(other, Reference):
             return isinstance(self, other.document_cls) and self.pk == other.pk
         return NotImplemented
 
@@ -190,8 +242,8 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
                      has already been commited to database given ``_id``
                      field could be generated before insertion
         """
-        value = self._data.get_by_mongo_name('_id')
-        return value if value is not missing else None
+        value = self._data.get(self.pk_field)
+        return value if value is not ma.missing else None
 
     @property
     def dbref(self):
@@ -204,7 +256,7 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
         return DBRef(collection=self.collection.name, id=self.pk)
 
     @classmethod
-    def build_from_mongo(cls, data, partial=False, use_cls=False):
+    def build_from_mongo(cls, data, use_cls=False):
         """
         Create a document instance from MongoDB data
 
@@ -216,17 +268,16 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
         if use_cls and '_cls' in data:
             cls = cls.opts.instance.retrieve_document(data['_cls'])
         doc = cls()
-        doc.from_mongo(data, partial=partial)
+        doc.from_mongo(data)
         return doc
 
-    def from_mongo(self, data, partial=False):
+    def from_mongo(self, data):
         """
         Update the document with the MongoDB data
 
         :param data: data as retrieved from MongoDB
         """
-        # TODO: handle partial
-        self._data.from_mongo(data, partial=partial)
+        self._data.from_mongo(data)
         self.is_created = True
 
     def to_mongo(self, update=False):
@@ -237,14 +288,13 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
                        update payload instead of containing the entire document
         """
         if update and not self.is_created:
-            raise NotCreatedError('Must create the document before'
-                                  ' using update')
+            raise NotCreatedError('Must create the document before using update')
         return self._data.to_mongo(update=update)
 
     def update(self, data):
-        """
-        Update the document with the given data.
-        """
+        """Update the document with the given data."""
+        if self.is_created and self.pk_field in data.keys():
+            raise AlreadyCreatedError("Can't modify id of a created document")
         self._data.update(data)
 
     def dump(self):
@@ -253,60 +303,39 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
         """
         return self._data.dump()
 
-    def clear_modified(self):
-        """
-        Reset the list of document's modified items.
-        """
-        self._data.clear_modified()
-
     def is_modified(self):
         """
         Returns True if and only if the document was modified since last commit.
         """
         return not self.is_created or self._data.is_modified()
 
-    def required_validate(self):
-        self._data.required_validate()
-
-    def items(self):
-        return self._data.items()
-
     # Data-proxy accessor shortcuts
 
-    def __getitem__(self, name):
-        value = self._data.get(name)
-        return value if value is not missing else None
+    def __setitem__(self, name, value):
+        if self.is_created and name == self.pk_field:
+            raise AlreadyCreatedError("Can't modify id of a created document")
+        super().__setitem__(name, value)
 
     def __delitem__(self, name):
-        self._data.delete(name)
-
-    def __setitem__(self, name, value):
-        self._data.set(name, value)
+        if self.is_created and name == self.pk_field:
+            raise AlreadyCreatedError("Can't modify id of a created document")
+        super().__delitem__(name)
 
     def __setattr__(self, name, value):
-        # Try to retrieve name among class's attributes and __slots__
-        if not self.__real_attributes:
-            # `dir(self)` result only depend on self's class so we can
-            # compute it once and store it inside the class
-            type(self).__real_attributes = dir(self)
-        if name in self.__real_attributes:
-            object.__setattr__(self, name, value)
+        if name in self._fields:
+            if self.is_created and name == self.pk_field:
+                raise AlreadyCreatedError("Can't modify id of a created document")
+            self._data.set(name, value)
         else:
-            self._data.set(name, value, to_raise=AttributeError)
-
-    def __getattr__(self, name):
-        if name[:2] == name[-2:] == '__':
-            raise AttributeError(name)
-        value = self._data.get(name, to_raise=AttributeError)
-        return value if value is not missing else None
+            super().__setattr__(name, value)
 
     def __delattr__(self, name):
-        if not self.__real_attributes:
-            type(self).__real_attributes = dir(self)
-        if name in self.__real_attributes:
-            object.__delattr__(self, name)
+        if name in self._fields:
+            if self.is_created and name == self.pk_field:
+                raise AlreadyCreatedError("Can't modify pk of a created document")
+            self._data.delete(name)
         else:
-            self._data.delete(name, to_raise=AttributeError)
+            super().__delattr__(name)
 
     # Callbacks
 
@@ -316,27 +345,24 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
 
     def pre_update(self):
         """
         Overload this method to get a callback before document update.
         :return: Additional filters dict that will be used for the query to
-            select the document to update.
+        select the document to update.
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
 
     def pre_delete(self):
         """
         Overload this method to get a callback before document deletion.
         :return: Additional filters dict that will be used for the query to
-            select the document to update.
+        select the document to update.
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
 
     def post_insert(self, ret):
         """
@@ -345,7 +371,6 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
 
     def post_update(self, ret):
         """
@@ -354,7 +379,6 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
 
     def post_delete(self, ret):
         """
@@ -363,4 +387,3 @@ class DocumentImplementation(BaseDataObject, Implementation, metaclass=MetaDocum
 
         .. note:: If you use an async driver, this callback can be asynchronous.
         """
-        pass
