@@ -1,7 +1,10 @@
-from .document import Implementation, Template
+"""umongo EmbeddedDocument"""
+import marshmallow as ma
+
+from .template import Implementation, Template
 from .data_objects import BaseDataObject
-from .data_proxy import missing
-from .exceptions import DocumentDefinitionError, AbstractDocumentError
+from .expose_missing import EXPOSE_MISSING
+from .exceptions import AbstractDocumentError
 
 
 __all__ = (
@@ -21,7 +24,6 @@ class EmbeddedDocumentTemplate(Template):
         :class:`umongo.instance.BaseInstance` to obtain it corresponding
         :class:`umongo.embedded_document.EmbeddedDocumentImplementation`.
     """
-    pass
 
 
 EmbeddedDocument = EmbeddedDocumentTemplate
@@ -32,7 +34,7 @@ class EmbeddedDocumentOpts:
     """
     Configuration for an :class:`umongo.embedded_document.EmbeddedDocument`.
 
-    Should be passed as a Meta class to the :class:`Document`
+    Should be passed as a Meta class to the :class:`EmbeddedDocument`
 
     .. code-block:: python
 
@@ -47,40 +49,34 @@ class EmbeddedDocumentOpts:
     ==================== ====================== ===========
     attribute            configurable in Meta   description
     ==================== ====================== ===========
-    template             no                     Origine template of the Document
+    template             no                     Origin template of the embedded document
     instance             no                     Implementation's instance
-    abstract             yes                    Document has no collection
-                                                and can only be inherited
-    allow_inheritance    yes                    Allow the document to be subclassed
-    is_child             no                     Document inherit of a non-abstract document
+    abstract             yes                    Embedded document can only be inherited
+    is_child             no                     Embedded document inherit of a non-abstract
+                                                embedded document
     strict               yes                    Don't accept unknown fields from mongo
                                                 (default: True)
-    offspring            no                     List of EmbeddedDocuments inheriting this one
+    offspring            no                     List of embedded documents inheriting this one
     ==================== ====================== ===========
     """
-
     def __repr__(self):
         return ('<{ClassName}('
                 'instance={self.instance}, '
                 'template={self.template}, '
                 'abstract={self.abstract}, '
-                'allow_inheritance={self.allow_inheritance}, '
                 'is_child={self.is_child}, '
                 'strict={self.strict}, '
                 'offspring={self.offspring})>'
                 .format(ClassName=self.__class__.__name__, self=self))
 
-    def __init__(self, instance, template, abstract=False, allow_inheritance=True,
+    def __init__(self, instance, template, abstract=False,
                  is_child=False, strict=True, offspring=None):
         self.instance = instance
         self.template = template
         self.abstract = abstract
-        self.allow_inheritance = allow_inheritance
         self.is_child = is_child
         self.strict = strict
         self.offspring = set(offspring) if offspring else set()
-        if self.abstract and not self.allow_inheritance:
-            raise DocumentDefinitionError("Abstract embedded document cannot disable inheritance")
 
 
 class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
@@ -89,8 +85,7 @@ class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
     :class:`umongo.instance.BaseInstance`.
     """
 
-    __slots__ = ('_callback', '_data', '_modified')
-    __real_attributes = None
+    __slots__ = ('_data', )
     opts = EmbeddedDocumentOpts(None, EmbeddedDocumentTemplate, abstract=True)
 
     def __init__(self, **kwargs):
@@ -106,7 +101,7 @@ class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
     def __eq__(self, other):
         if isinstance(other, dict):
             return self._data == other
-        elif hasattr(other, '_data'):
+        if hasattr(other, '_data'):
             return self._data == other._data
         return NotImplemented
 
@@ -114,6 +109,9 @@ class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
         return self._data.is_modified()
 
     def clear_modified(self):
+        """
+        Reset the list of document's modified items.
+        """
         self._data.clear_modified()
 
     def required_validate(self):
@@ -160,7 +158,7 @@ class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
 
     def __getitem__(self, name):
         value = self._data.get(name)
-        return value if value is not missing else None
+        return None if value is ma.missing and not EXPOSE_MISSING.get() else value
 
     def __delitem__(self, name):
         self._data.delete(name)
@@ -169,26 +167,19 @@ class EmbeddedDocumentImplementation(Implementation, BaseDataObject):
         self._data.set(name, value)
 
     def __setattr__(self, name, value):
-        # Try to retrieve name among class's attributes and __slots__
-        if not self.__real_attributes:
-            # `dir(self)` result only depend on self's class so we can
-            # compute it once and store it inside the class
-            type(self).__real_attributes = dir(self)
-        if name in self.__real_attributes:
-            object.__setattr__(self, name, value)
+        if name in self._fields:
+            self._data.set(name, value)
         else:
-            self._data.set(name, value, to_raise=AttributeError)
+            super().__setattr__(name, value)
 
     def __getattr__(self, name):
-        if name[:2] == name[-2:] == '__':
-            raise AttributeError(name)
-        value = self._data.get(name, to_raise=AttributeError)
-        return value if value is not missing else None
+        if name in self._fields:
+            value = self._data.get(name)
+            return None if value is ma.missing and not EXPOSE_MISSING.get() else value
+        raise AttributeError(name)
 
     def __delattr__(self, name):
-        if not self.__real_attributes:
-            type(self).__real_attributes = dir(self)
-        if name in self.__real_attributes:
-            object.__delattr__(self, name)
+        if name in self._fields:
+            self._data.delete(name)
         else:
-            self._data.delete(name, to_raise=AttributeError)
+            super().__delattr__(name)
