@@ -1,31 +1,34 @@
+"""umongo fields"""
+import collections
 import datetime as dt
 
 from bson import DBRef, ObjectId, Decimal128
-from marshmallow import ValidationError, missing
-from marshmallow import fields as ma_fields
+import marshmallow as ma
 
 # from .registerer import retrieve_document
-from .exceptions import NotRegisteredDocumentError
+from .document import DocumentImplementation
+from .exceptions import NotRegisteredDocumentError, DocumentDefinitionError
 from .template import get_template
 from .data_objects import Reference, List, Dict
 from . import marshmallow_bonus as ma_bonus_fields
-from .abstract import BaseField
+from .abstract import BaseField, I18nErrorDict
 from .i18n import gettext as _
 
 
 __all__ = (
     # 'RawField',
-    'DictField',
-    'ListField',
+    # 'MappingField',
+    # 'TupleField',
     'StringField',
     'UUIDField',
     'NumberField',
     'IntegerField',
     'DecimalField',
     'BooleanField',
-    'FormattedStringField',
     'FloatField',
     'DateTimeField',
+    'NaiveDateTimeField',
+    'AwareDateTimeField',
     # 'TimeField',
     'DateField',
     # 'TimeDeltaField',
@@ -35,8 +38,10 @@ __all__ = (
     'StrField',
     'BoolField',
     'IntField',
+    'DictField',
+    'ListField',
     'ConstantField',
-    'StrictDateTimeField',
+    # 'PluckField'
     'ObjectIdField',
     'ReferenceField',
     'GenericReferenceField',
@@ -47,105 +52,27 @@ __all__ = (
 # Republish supported marshmallow fields
 
 
-# class RawField(BaseField, ma_fields.Raw):
+# class RawField(BaseField, ma.fields.Raw):
 #     pass
 
 
-class DictField(BaseField, ma_fields.Dict):
-
-    def _deserialize(self, value, attr, data):
-        value = super()._deserialize(value, attr, data)
-        return Dict(value)
-
-    def _serialize_to_mongo(self, obj):
-        if obj is None:
-            return missing
-        return dict(obj)
-
-    def _deserialize_from_mongo(self, value):
-        if value:
-            return Dict(value)
-        return Dict()
-
-    def translate_query(self, key, query):
-        keys = key.split('.')
-        self.attribute or keys[0] + '.' + '.'.join(keys[1:])
-        return {self.attribute or key: query}
-
-
-class ListField(BaseField, ma_fields.List):
-
-    def _deserialize(self, value, attr, data):
-        return List(self.container, super()._deserialize(value, attr, data))
-
-    def _serialize_to_mongo(self, obj):
-        if obj is None:
-            return missing
-        return [self.container.serialize_to_mongo(each) for each in obj]
-
-    def _deserialize_from_mongo(self, value):
-        if value:
-            return List(self.container, [self.container.deserialize_from_mongo(each)
-                                         for each in value])
-        return List(self.container)
-
-    def map_to_field(self, mongo_path, path, func):
-        """Apply a function to every field in the schema
-        """
-        func(mongo_path, path, self.container)
-        if hasattr(self.container, 'map_to_field'):
-            self.container.map_to_field(mongo_path, path, func)
-
-    def as_marshmallow_field(self, params=None, mongo_world=False, **kwargs):
-        # Overwrite default `as_marshmallow_field` to handle deserialization
-        # difference (`_id` vs `id`)
-        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
-        if params:
-            container_params = params.pop('params', None)
-            field_kwargs.update(params)
-        else:
-            container_params = None
-        container_ma_schema = self.container.as_marshmallow_field(
-            mongo_world=mongo_world, params=container_params, **kwargs)
-        return ma_fields.List(container_ma_schema, **field_kwargs)
-
-    def _required_validate(self, value):
-        if value is missing or not hasattr(self.container, '_required_validate'):
-            return
-        required_validate = self.container._required_validate
-        errors = {}
-        for i, sub_value in enumerate(value):
-            try:
-                required_validate(sub_value)
-            except ValidationError as exc:
-                errors[i] = exc.messages
-        if errors:
-            raise ValidationError(errors)
-
-
-class StringField(BaseField, ma_fields.String):
+class StringField(BaseField, ma.fields.String):
     pass
 
 
-class UUIDField(BaseField, ma_fields.UUID):
+class UUIDField(BaseField, ma.fields.UUID):
     pass
 
 
-class NumberField(BaseField, ma_fields.Number):
+class NumberField(BaseField, ma.fields.Number):
     pass
 
 
-class IntegerField(BaseField, ma_fields.Integer):
+class IntegerField(BaseField, ma.fields.Integer):
     pass
 
 
-class DecimalField(BaseField, ma_fields.Decimal):
-    """
-    .. warning::
-
-        This field is only supported on MongoDB 3.4+.
-        Using it on older versions will result in uncaught errors.
-    """
+class DecimalField(BaseField, ma.fields.Decimal):
     def _serialize_to_mongo(self, obj):
         return Decimal128(obj)
 
@@ -153,15 +80,11 @@ class DecimalField(BaseField, ma_fields.Decimal):
         return value.to_decimal()
 
 
-class BooleanField(BaseField, ma_fields.Boolean):
+class BooleanField(BaseField, ma.fields.Boolean):
     pass
 
 
-class FormattedStringField(BaseField, ma_fields.FormattedString):
-    pass
-
-
-class FloatField(BaseField, ma_fields.Float):
+class FloatField(BaseField, ma.fields.Float):
     pass
 
 
@@ -177,34 +100,50 @@ def _round_to_millisecond(datetime):
     return datetime.replace(microsecond=microseconds)
 
 
-class DateTimeField(BaseField, ma_fields.DateTime):
+class DateTimeField(BaseField, ma.fields.DateTime):
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         if isinstance(value, dt.datetime):
             ret = value
         else:
-            ret = super()._deserialize(value, attr, data)
+            ret = super()._deserialize(value, attr, data, **kwargs)
         return _round_to_millisecond(ret)
 
 
-class LocalDateTimeField(BaseField, ma_fields.LocalDateTime):
+class NaiveDateTimeField(BaseField, ma.fields.NaiveDateTime):
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         if isinstance(value, dt.datetime):
             ret = value
         else:
-            ret = super()._deserialize(value, attr, data)
+            ret = super()._deserialize(value, attr, data, **kwargs)
         return _round_to_millisecond(ret)
 
 
-# class TimeField(BaseField, ma_fields.Time):
+class AwareDateTimeField(BaseField, ma.fields.AwareDateTime):
+
+    def _deserialize(self, value, attr, data, **kwargs):
+        if isinstance(value, dt.datetime):
+            ret = value
+        else:
+            ret = super()._deserialize(value, attr, data, **kwargs)
+        return _round_to_millisecond(ret)
+
+    def _deserialize_from_mongo(self, value):
+        value = value.replace(tzinfo=dt.timezone.utc)
+        if self.default_timezone is not None:
+            value = value.astimezone(self.default_timezone)
+        return value
+
+
+# class TimeField(BaseField, ma.fields.Time):
 #     pass
 
 
-class DateField(BaseField, ma_fields.Date):
+class DateField(BaseField, ma.fields.Date):
     """This field converts a date to a datetime to store it as a BSON Date"""
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         if isinstance(value, dt.date):
             return value
         return super()._deserialize(value, attr, data)
@@ -216,22 +155,148 @@ class DateField(BaseField, ma_fields.Date):
         return value.date()
 
 
-# class TimeDeltaField(BaseField, ma_fields.TimeDelta):
+# class TimeDeltaField(BaseField, ma.fields.TimeDelta):
 #     pass
 
 
-class UrlField(BaseField, ma_fields.Url):
+class UrlField(BaseField, ma.fields.Url):
     pass
 
 
-class EmailField(BaseField, ma_fields.Email):
+class EmailField(BaseField, ma.fields.Email):
     pass
 
 
-class ConstantField(BaseField, ma_fields.Constant):
+class ConstantField(BaseField, ma.fields.Constant):
     pass
 
 
+class DictField(BaseField, ma.fields.Dict):
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        def cast_value_or_callable(key_field, value_field, value):
+            if value is ma.missing:
+                return ma.missing
+            if callable(value):
+                return lambda: Dict(key_field, value_field, value())
+            return Dict(key_field, value_field, value)
+
+        self.default = cast_value_or_callable(self.key_field, self.value_field, self.default)
+        self.missing = cast_value_or_callable(self.key_field, self.value_field, self.missing)
+
+    def _deserialize(self, value, attr, data, **kwargs):
+        value = super()._deserialize(value, attr, data, **kwargs)
+        return Dict(self.key_field, self.value_field, value)
+
+    def _serialize_to_mongo(self, obj):
+        if obj is None:
+            return ma.missing
+        return {
+            self.key_field.serialize_to_mongo(k) if self.key_field else k:
+            self.value_field.serialize_to_mongo(v) if self.value_field else v
+            for k, v in obj.items()
+        }
+
+    def _deserialize_from_mongo(self, value):
+        if value:
+            return Dict(
+                self.key_field,
+                self.value_field,
+                {
+                    self.key_field.deserialize_from_mongo(k) if self.key_field else k:
+                    self.value_field.deserialize_from_mongo(v) if self.value_field else v
+                    for k, v in value.items()
+                }
+            )
+        return Dict(self.key_field, self.value_field)
+
+    def as_marshmallow_field(self):
+        field_kwargs = self._extract_marshmallow_field_params()
+        if self.value_field:
+            inner_ma_field = self.value_field.as_marshmallow_field()
+        else:
+            inner_ma_field = None
+        m_field = ma.fields.Dict(
+            self.key_field, inner_ma_field, metadata=self.metadata, **field_kwargs)
+        m_field.error_messages = I18nErrorDict(m_field.error_messages)
+        return m_field
+
+    def _required_validate(self, value):
+        if not hasattr(self.value_field, '_required_validate'):
+            return
+        required_validate = self.value_field._required_validate
+        errors = collections.defaultdict(dict)
+        for key, val in value.items():
+            try:
+                required_validate(val)
+            except ma.ValidationError as exc:
+                errors[key]["value"] = exc.messages
+        if errors:
+            raise ma.ValidationError(errors)
+
+
+class ListField(BaseField, ma.fields.List):
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        def cast_value_or_callable(inner, value):
+            if value is ma.missing:
+                return ma.missing
+            if callable(value):
+                return lambda: List(inner, value())
+            return List(inner, value)
+
+        self.default = cast_value_or_callable(self.inner, self.default)
+        self.missing = cast_value_or_callable(self.inner, self.missing)
+
+    def _deserialize(self, value, attr, data, **kwargs):
+        value = super()._deserialize(value, attr, data, **kwargs)
+        return List(self.inner, value)
+
+    def _serialize_to_mongo(self, obj):
+        if obj is None:
+            return ma.missing
+        return [self.inner.serialize_to_mongo(each) for each in obj]
+
+    def _deserialize_from_mongo(self, value):
+        if value:
+            return List(
+                self.inner,
+                [self.inner.deserialize_from_mongo(each) for each in value]
+            )
+        return List(self.inner)
+
+    def map_to_field(self, mongo_path, path, func):
+        """Apply a function to every field in the schema
+        """
+        func(mongo_path, path, self.inner)
+        if hasattr(self.inner, 'map_to_field'):
+            self.inner.map_to_field(mongo_path, path, func)
+
+    def as_marshmallow_field(self):
+        field_kwargs = self._extract_marshmallow_field_params()
+        inner_ma_field = self.inner.as_marshmallow_field()
+        m_field = ma.fields.List(inner_ma_field, metadata=self.metadata, **field_kwargs)
+        m_field.error_messages = I18nErrorDict(m_field.error_messages)
+        return m_field
+
+    def _required_validate(self, value):
+        if not hasattr(self.inner, '_required_validate'):
+            return
+        required_validate = self.inner._required_validate
+        errors = {}
+        for i, sub_value in enumerate(value):
+            try:
+                required_validate(sub_value)
+            except ma.ValidationError as exc:
+                errors[i] = exc.messages
+        if errors:
+            raise ma.ValidationError(errors)
+
+
 # Aliases
 URLField = UrlField
 StrField = StringField
@@ -239,21 +304,6 @@ BoolField = BooleanField
 IntField = IntegerField
 
 
-# Bonus: new fields !
-
-class StrictDateTimeField(BaseField, ma_bonus_fields.StrictDateTime):
-
-    def _deserialize(self, value, attr, data):
-        if isinstance(value, dt.datetime):
-            ret = self._set_tz_awareness(value)
-        else:
-            ret = super()._deserialize(value, attr, data)
-        return _round_to_millisecond(ret)
-
-    def _deserialize_from_mongo(self, value):
-        return self._set_tz_awareness(value)
-
-
 class ObjectIdField(BaseField, ma_bonus_fields.ObjectId):
     pass
 
@@ -265,6 +315,8 @@ class ReferenceField(BaseField, ma_bonus_fields.Reference):
         :param document: Can be a :class:`umongo.embedded_document.DocumentTemplate`,
             another instance's :class:`umongo.embedded_document.DocumentImplementation` or
             the embedded document class name.
+
+        .. warning:: The referenced document's _id must be an `ObjectId`.
         """
         super().__init__(*args, **kwargs)
         # TODO : check document_cls is implementation or string
@@ -275,8 +327,6 @@ class ReferenceField(BaseField, ma_bonus_fields.Reference):
         else:
             self.document = document
         self._document_cls = None
-        # Avoid importing multiple times
-        from .document import DocumentImplementation
         self._document_implementation_cls = DocumentImplementation
 
     @property
@@ -289,30 +339,30 @@ class ReferenceField(BaseField, ma_bonus_fields.Reference):
             self._document_cls = self.instance.retrieve_document(self.document)
         return self._document_cls
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         if value is None:
             return None
         if isinstance(value, DBRef):
-            if self._document_cls.collection.name != value.collection:
-                raise ValidationError(_("DBRef must be on collection `{collection}`.").format(
-                    self._document_cls.collection.name))
+            if self.document_cls.collection.name != value.collection:
+                raise ma.ValidationError(_("DBRef must be on collection `{collection}`.").format(
+                    self.document_cls.collection.name))
             value = value.id
         elif isinstance(value, Reference):
             if value.document_cls != self.document_cls:
-                raise ValidationError(_("`{document}` reference expected.").format(
+                raise ma.ValidationError(_("`{document}` reference expected.").format(
                     document=self.document_cls.__name__))
-            if type(value) is not self.reference_cls:
+            if not isinstance(value, self.reference_cls):
                 value = self.reference_cls(value.document_cls, value.pk)
             return value
         elif isinstance(value, self.document_cls):
             if not value.is_created:
-                raise ValidationError(
+                raise ma.ValidationError(
                     _("Cannot reference a document that has not been created yet."))
             value = value.pk
         elif isinstance(value, self._document_implementation_cls):
-            raise ValidationError(_("`{document}` reference expected.").format(
+            raise ma.ValidationError(_("`{document}` reference expected.").format(
                 document=self.document_cls.__name__))
-        value = super()._deserialize(value, attr, data)
+        value = super()._deserialize(value, attr, data, **kwargs)
         return self.reference_cls(self.document_cls, value)
 
     def _serialize_to_mongo(self, obj):
@@ -321,29 +371,19 @@ class ReferenceField(BaseField, ma_bonus_fields.Reference):
     def _deserialize_from_mongo(self, value):
         return self.reference_cls(self.document_cls, value)
 
-    def as_marshmallow_field(self, params=None, mongo_world=False, **kwargs):
-        # Overwrite default `as_marshmallow_field` to handle deserialization
-        # difference (`_id` vs `id`)
-        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
-        if params:
-            field_kwargs.update(params)
-        return ma_bonus_fields.Reference(mongo_world=mongo_world, **field_kwargs)
-
 
 class GenericReferenceField(BaseField, ma_bonus_fields.GenericReference):
 
     def __init__(self, *args, reference_cls=Reference, **kwargs):
         super().__init__(*args, **kwargs)
         self.reference_cls = reference_cls
-        # Avoid importing multiple times
-        from .document import DocumentImplementation
         self._document_implementation_cls = DocumentImplementation
 
     def _document_cls(self, class_name):
         try:
             return self.instance.retrieve_document(class_name)
         except NotRegisteredDocumentError:
-            raise ValidationError(_('Unknown document `{document}`.').format(
+            raise ma.ValidationError(_('Unknown document `{document}`.').format(
                 document=class_name))
 
     def _serialize(self, value, attr, obj):
@@ -351,29 +391,28 @@ class GenericReferenceField(BaseField, ma_bonus_fields.GenericReference):
             return None
         return {'id': str(value.pk), 'cls': value.document_cls.__name__}
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         if value is None:
             return None
         if isinstance(value, Reference):
-            if type(value) is not self.reference_cls:
+            if not isinstance(value, self.reference_cls):
                 value = self.reference_cls(value.document_cls, value.pk)
             return value
-        elif isinstance(value, self._document_implementation_cls):
+        if isinstance(value, self._document_implementation_cls):
             if not value.is_created:
-                raise ValidationError(
+                raise ma.ValidationError(
                     _("Cannot reference a document that has not been created yet."))
             return self.reference_cls(value.__class__, value.pk)
-        elif isinstance(value, dict):
+        if isinstance(value, dict):
             if value.keys() != {'cls', 'id'}:
-                raise ValidationError(_("Generic reference must have `id` and `cls` fields."))
+                raise ma.ValidationError(_("Generic reference must have `id` and `cls` fields."))
             try:
                 _id = ObjectId(value['id'])
             except ValueError:
-                raise ValidationError(_("Invalid `id` field."))
+                raise ma.ValidationError(_("Invalid `id` field."))
             document_cls = self._document_cls(value['cls'])
             return self.reference_cls(document_cls, _id)
-        else:
-            raise ValidationError(_("Invalid value for generic reference field."))
+        raise ma.ValidationError(_("Invalid value for generic reference field."))
 
     def _serialize_to_mongo(self, obj):
         return {'_id': obj.pk, '_cls': obj.document_cls.__name__}
@@ -382,16 +421,8 @@ class GenericReferenceField(BaseField, ma_bonus_fields.GenericReference):
         document_cls = self._document_cls(value['_cls'])
         return self.reference_cls(document_cls, value['_id'])
 
-    def as_marshmallow_field(self, params=None, mongo_world=False, **kwargs):
-        # Overwrite default `as_marshmallow_field` to handle deserialization
-        # difference (`_id` vs `id`)
-        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
-        if params:
-            field_kwargs.update(params)
-        return ma_bonus_fields.GenericReference(mongo_world=mongo_world, **field_kwargs)
 
-
-class EmbeddedField(BaseField, ma_fields.Nested):
+class EmbeddedField(BaseField, ma.fields.Nested):
 
     def __init__(self, embedded_document, *args, **kwargs):
         """
@@ -425,8 +456,12 @@ class EmbeddedField(BaseField, ma_fields.Nested):
         implementing the `embedded_document` attribute.
         """
         if not self._embedded_document_cls:
-            self._embedded_document_cls = self.instance.retrieve_embedded_document(
-                self.embedded_document)
+            embedded_document_cls = self.instance.retrieve_embedded_document(self.embedded_document)
+            if embedded_document_cls.opts.abstract:
+                raise DocumentDefinitionError(
+                    "EmbeddedField doesn't accept abstract embedded document"
+                )
+            self._embedded_document_cls = embedded_document_cls
         return self._embedded_document_cls
 
     def _serialize(self, value, attr, obj):
@@ -434,27 +469,26 @@ class EmbeddedField(BaseField, ma_fields.Nested):
             return None
         return value.dump()
 
-    def _deserialize(self, value, attr, data):
+    def _deserialize(self, value, attr, data, **kwargs):
         embedded_document_cls = self.embedded_document_cls
         if isinstance(value, embedded_document_cls):
             return value
         if not isinstance(value, dict):
-            raise ValidationError({'_schema': ['Invalid input type.']})
+            raise ma.ValidationError({'_schema': ['Invalid input type.']})
         # Handle inheritance deserialization here using `cls` field as hint
         if embedded_document_cls.opts.offspring and 'cls' in value:
             to_use_cls_name = value.pop('cls')
             if not any(o for o in embedded_document_cls.opts.offspring
                        if o.__name__ == to_use_cls_name):
-                raise ValidationError(_('Unknown document `{document}`.').format(
+                raise ma.ValidationError(_('Unknown document `{document}`.').format(
                     document=to_use_cls_name))
             try:
                 to_use_cls = embedded_document_cls.opts.instance.retrieve_embedded_document(
                     to_use_cls_name)
-            except NotRegisteredDocumentError as e:
-                raise ValidationError(str(e))
+            except NotRegisteredDocumentError as exc:
+                raise ma.ValidationError(str(exc))
             return to_use_cls(**value)
-        else:
-            return embedded_document_cls(**value)
+        return embedded_document_cls(**value)
 
     def _serialize_to_mongo(self, obj):
         return obj.to_mongo()
@@ -466,13 +500,13 @@ class EmbeddedField(BaseField, ma_fields.Nested):
         # Overload default to handle recursive check
         super()._validate_missing(value)
         errors = {}
-        if value is missing:
-            def get_sub_value(key):
-                return missing
+        if value is ma.missing:
+            def get_sub_value(_):
+                return ma.missing
         elif isinstance(value, dict):
             # value is a dict for deserialization
             def get_sub_value(key):
-                return value.get(key, missing)
+                return value.get(key, ma.missing)
         elif isinstance(value, self.embedded_document_cls):
             # value is a valid EmbeddedDocument
             def get_sub_value(key):
@@ -485,14 +519,14 @@ class EmbeddedField(BaseField, ma_fields.Nested):
             sub_value = get_sub_value(name)
             # `_validate_missing` doesn't check for required fields here, so we
             # can safely skip missing values
-            if sub_value is missing:
+            if sub_value is ma.missing:
                 continue
             try:
                 field._validate_missing(sub_value)
-            except ValidationError as ve:
-                errors[name] = ve.messages
+            except ma.ValidationError as exc:
+                errors[name] = exc.messages
         if errors:
-            raise ValidationError(errors)
+            raise ma.ValidationError(errors)
 
     def map_to_field(self, mongo_path, path, func):
         """Apply a function to every field in the schema"""
@@ -503,22 +537,13 @@ class EmbeddedField(BaseField, ma_fields.Nested):
             if hasattr(field, 'map_to_field'):
                 field.map_to_field(cur_mongo_path, cur_path, func)
 
-    def as_marshmallow_field(self, params=None, mongo_world=False, **kwargs):
+    def as_marshmallow_field(self):
         # Overwrite default `as_marshmallow_field` to handle nesting
-        field_kwargs = self._extract_marshmallow_field_params(mongo_world)
-        if params:
-            nested_params = params.pop('params', None)
-            nested_meta = params.pop('meta', None)
-            field_kwargs.update(params)
-        else:
-            nested_params = None
-            nested_meta = None
-        schema_kwargs = {k: v for k, v in kwargs.items()
-                         if k in ('base_schema_cls', 'check_unknown_fields')}
-        nested_ma_schema = self._embedded_document_cls.schema.as_marshmallow_schema(
-            params=nested_params, mongo_world=mongo_world, meta=nested_meta, **schema_kwargs)
-        return ma_fields.Nested(nested_ma_schema, **field_kwargs)
+        field_kwargs = self._extract_marshmallow_field_params()
+        nested_ma_schema = self.embedded_document_cls.schema.as_marshmallow_schema()
+        m_field = ma.fields.Nested(nested_ma_schema, metadata=self.metadata, **field_kwargs)
+        m_field.error_messages = I18nErrorDict(m_field.error_messages)
+        return m_field
 
     def _required_validate(self, value):
-        if value is not missing:
-            value.required_validate()
+        value.required_validate()
