@@ -1,16 +1,24 @@
+import collections
+from contextvars import ContextVar
+from contextlib import contextmanager
+
 from pymongo.database import Database
 from pymongo.cursor import Cursor
 from pymongo.errors import DuplicateKeyError
+import marshmallow as ma
 
 from ..builder import BaseBuilder
+from ..instance import Instance
 from ..document import DocumentImplementation
-from ..data_proxy import missing
 from ..data_objects import Reference
-from ..exceptions import NotCreatedError, UpdateError, DeleteError, ValidationError
-from ..fields import ReferenceField, ListField, EmbeddedField
+from ..exceptions import NotCreatedError, UpdateError, DeleteError, NoneReferenceError
+from ..fields import ReferenceField, ListField, DictField, EmbeddedField
 from ..query_mapper import map_query
 
-from .tools import cook_find_filter
+from .tools import cook_find_filter, remove_cls_field_from_embedded_docs
+
+
+SESSION = ContextVar("session", default=None)
 
 
 # pymongo.Cursor defines __del__ method, hence mongomock's WrappedCursor should
@@ -21,7 +29,7 @@ class BaseWrappedCursor:
 
     def __init__(self, document_cls, cursor, *args, **kwargs):
         # Such a cunning plan my lord !
-        # We inherit from Cursor but don't call it __init__ because
+        # We inherit from Cursor but don't call its __init__ because
         # we act as a proxy to the underlying raw_cursor
         WrappedCursor.raw_cursor.__set__(self, cursor)
         WrappedCursor.document_cls.__set__(self, document_cls)
@@ -69,13 +77,13 @@ class PyMongoDocument(DocumentImplementation):
         """
         if not self.is_created:
             raise NotCreatedError("Document doesn't exists in database")
-        ret = self.collection.find_one(self.pk)
+        ret = self.collection.find_one(self.pk, session=SESSION.get())
         if ret is None:
             raise NotCreatedError("Document doesn't exists in database")
         self._data = self.DataProxy()
         self._data.from_mongo(ret)
 
-    def commit(self, io_validate_all=False, conditions=None):
+    def commit(self, io_validate_all=False, conditions=None, replace=False):
         """
         Commit the document in database.
         If the document doesn't already exist it will be inserted, otherwise
@@ -86,12 +94,13 @@ class PyMongoDocument(DocumentImplementation):
             satisfies condition(s) (e.g. version number).
             Raises :class:`umongo.exceptions.UpdateError` if the
             conditions are not satisfied.
+        :param replace: Replace the document rather than update.
         :return: A :class:`pymongo.results.UpdateResult` or
             :class:`pymongo.results.InsertOneResult` depending of the operation.
         """
         try:
             if self.is_created:
-                if self.is_modified():
+                if self.is_modified() or replace:
                     query = conditions or {}
                     query['_id'] = self.pk
                     # pre_update can provide additional query filter and/or
@@ -101,43 +110,46 @@ class PyMongoDocument(DocumentImplementation):
                         query.update(map_query(additional_filter, self.schema.fields))
                     self.required_validate()
                     self.io_validate(validate_all=io_validate_all)
-                    payload = self._data.to_mongo(update=True)
-                    ret = self.collection.update_one(query, payload)
+                    if replace:
+                        payload = self._data.to_mongo(update=False)
+                        ret = self.collection.replace_one(query, payload, session=SESSION.get())
+                    else:
+                        payload = self._data.to_mongo(update=True)
+                        ret = self.collection.update_one(query, payload, session=SESSION.get())
                     if ret.matched_count != 1:
                         raise UpdateError(ret)
                     self.post_update(ret)
                 else:
                     ret = None
             elif conditions:
-                raise RuntimeError('Document must already exist in database to use `conditions`.')
+                raise NotCreatedError(
+                    'Document must already exist in database to use `conditions`.'
+                )
             else:
                 self.pre_insert()
                 self.required_validate()
                 self.io_validate(validate_all=io_validate_all)
                 payload = self._data.to_mongo(update=False)
-                ret = self.collection.insert_one(payload)
+                ret = self.collection.insert_one(payload, session=SESSION.get())
                 # TODO: check ret ?
-                self._data.set_by_mongo_name('_id', ret.inserted_id)
+                self._data.set(self.pk_field, ret.inserted_id)
                 self.is_created = True
                 self.post_insert(ret)
         except DuplicateKeyError as exc:
-            # Need to dig into error message to find faulting index
-            errmsg = exc.details['errmsg']
-            for index in self.opts.indexes:
-                if ('.$%s' % index.document['name'] in errmsg or
-                        ' %s ' % index.document['name'] in errmsg):
-                    keys = index.document['key'].keys()
-                    if len(keys) == 1:
-                        msg = self.schema.fields[keys[0]].error_messages['unique']
-                        raise ValidationError({keys[0]: msg})
-                    else:
-                        fields = self.schema.fields
-                        # Compound index (sort value to make testing easier)
-                        keys = sorted(keys)
-                        raise ValidationError({k: fields[k].error_messages[
-                            'unique_compound'].format(fields=keys) for k in keys})
-            # Unknown index, cannot wrap the error so just reraise it
-            raise
+            # Sort value to make testing easier for compound indexes
+            keys = sorted(exc.details['keyPattern'].keys())
+            try:
+                fields = [self.schema.fields[k] for k in keys]
+            except KeyError:
+                # A key in the index is unknwon from umongo
+                raise exc
+            if len(keys) == 1:
+                msg = fields[0].error_messages['unique']
+                raise ma.ValidationError({keys[0]: msg})
+            raise ma.ValidationError({
+                k: f.error_messages['unique_compound'].format(fields=keys)
+                for k, f in zip(keys, fields)
+            })
         self._data.clear_modified()
         return ret
 
@@ -164,7 +176,7 @@ class PyMongoDocument(DocumentImplementation):
         additional_filter = self.pre_delete()
         if additional_filter:
             query.update(map_query(additional_filter, self.schema.fields))
-        ret = self.collection.delete_one(query)
+        ret = self.collection.delete_one(query, session=SESSION.get())
         if ret.deleted_count != 1:
             raise DeleteError(ret)
         self.is_created = False
@@ -190,7 +202,7 @@ class PyMongoDocument(DocumentImplementation):
         Find a single document in database.
         """
         filter = cook_find_filter(cls, filter)
-        ret = cls.collection.find_one(filter, *args, **kwargs)
+        ret = cls.collection.find_one(filter, session=SESSION.get(), *args, **kwargs)
         if ret is not None:
             ret = cls.build_from_mongo(ret, use_cls=True)
         return ret
@@ -203,7 +215,7 @@ class PyMongoDocument(DocumentImplementation):
         Returns a cursor that provide Documents.
         """
         filter = cook_find_filter(cls, filter)
-        raw_cursor = cls.collection.find(filter, *args, **kwargs)
+        raw_cursor = cls.collection.find(filter, session=SESSION.get(), *args, **kwargs)
         return cls.cursor_cls(cls, raw_cursor)
 
     @classmethod
@@ -215,15 +227,15 @@ class PyMongoDocument(DocumentImplementation):
         defaults to an empty filter.
         """
         filter = cook_find_filter(cls, filter or {})
-        return cls.collection.count_documents(filter, **kwargs)
+        return cls.collection.count_documents(filter, session=SESSION.get(), **kwargs)
 
     @classmethod
     def ensure_indexes(cls):
         """
         Check&create if needed the Document's indexes in database
         """
-        if cls.opts.indexes:
-            cls.collection.create_indexes(cls.opts.indexes)
+        if cls.indexes:
+            cls.collection.create_indexes(cls.indexes, session=SESSION.get())
 
 
 # Run multiple validators and collect all errors in one
@@ -235,10 +247,10 @@ def _run_validators(validators, field, value):
         for validator in validators:
             try:
                 validator(field, value)
-            except ValidationError as ve:
-                errors.extend(ve.messages)
+            except ma.ValidationError as exc:
+                errors.extend(exc.messages)
         if errors:
-            raise ValidationError(errors)
+            raise ma.ValidationError(errors)
 
 
 def _io_validate_data_proxy(schema, data_proxy, partial=None):
@@ -246,40 +258,61 @@ def _io_validate_data_proxy(schema, data_proxy, partial=None):
     for name, field in schema.fields.items():
         if partial and name not in partial:
             continue
-        data_name = field.attribute or name
-        value = data_proxy._data[data_name]
-        if value is missing:
+        value = data_proxy.get(name)
+        if value is ma.missing:
             continue
         try:
             if field.io_validate_recursive:
                 field.io_validate_recursive(field, value)
             if field.io_validate:
                 _run_validators(field.io_validate, field, value)
-        except ValidationError as ve:
-            errors[name] = ve.messages
+        except ma.ValidationError as exc:
+            errors[name] = exc.messages
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
 
 
 def _reference_io_validate(field, value):
+    if value is None:
+        return
     value.fetch(no_data=True)
 
 
 def _list_io_validate(field, value):
+    if not value:
+        return
     errors = {}
-    validators = field.container.io_validate
+    validators = field.inner.io_validate
     if not validators:
         return
-    for i, e in enumerate(value):
+    for idx, val in enumerate(value):
         try:
-            _run_validators(validators, field.container, e)
-        except ValidationError as ev:
-            errors[i] = ev.messages
+            _run_validators(validators, field.inner, val)
+        except ma.ValidationError as exc:
+            errors[idx] = exc.messages
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
+
+
+def _dict_io_validate(field, value):
+    if not value or not field.value_field:
+        return
+    errors = collections.defaultdict(dict)
+    validators = field.value_field.io_validate
+    if not validators:
+        return
+    for key, val in value.items():
+        try:
+            _run_validators(validators, field.value_field, val)
+        except ma.ValidationError as exc:
+            errors[key]["value"] = exc.messages
+    if errors:
+        raise ma.ValidationError(errors)
 
 
 def _embedded_document_io_validate(field, value):
+    if not value:
+        return
     _io_validate_data_proxy(value.schema, value._data)
 
 
@@ -292,10 +325,10 @@ class PyMongoReference(Reference):
     def fetch(self, no_data=False, force_reload=False):
         if not self._document or force_reload:
             if self.pk is None:
-                raise ReferenceError('Cannot retrieve a None Reference')
+                raise NoneReferenceError('Cannot retrieve a None Reference')
             self._document = self.document_cls.find_one(self.pk)
             if not self._document:
-                raise ValidationError(self.error_messages['not_found'].format(
+                raise ma.ValidationError(self.error_messages['not_found'].format(
                     document=self.document_cls.__name__))
         return self._document
 
@@ -304,10 +337,6 @@ class PyMongoBuilder(BaseBuilder):
 
     BASE_DOCUMENT_CLS = PyMongoDocument
 
-    @staticmethod
-    def is_compatible_with(db):
-        return isinstance(db, Database)
-
     def _patch_field(self, field):
         super()._patch_field(field)
 
@@ -321,8 +350,55 @@ class PyMongoBuilder(BaseBuilder):
                 field.io_validate = [validators]
         if isinstance(field, ListField):
             field.io_validate_recursive = _list_io_validate
+        if isinstance(field, DictField):
+            field.io_validate_recursive = _dict_io_validate
         if isinstance(field, ReferenceField):
             field.io_validate.append(_reference_io_validate)
             field.reference_cls = PyMongoReference
         if isinstance(field, EmbeddedField):
             field.io_validate_recursive = _embedded_document_io_validate
+
+
+class PyMongoInstance(Instance):
+    """
+    :class:`umongo.instance.Instance` implementation for pymongo
+    """
+    BUILDER_CLS = PyMongoBuilder
+
+    @staticmethod
+    def is_compatible_with(db):
+        return isinstance(db, Database)
+
+    @contextmanager
+    def session(self):
+        with self.db.client.start_session() as session:
+            try:
+                token = SESSION.set(session)
+                yield session
+            finally:
+                SESSION.reset(token)
+
+
+class PyMongoMigrationInstance(PyMongoInstance):
+    """PyMongo instance with migration features"""
+
+    def migrate_2_to_3(self):
+        """Migrate database from umongo 2 to umongo 3
+
+        - EmbeddedDocument _cls field is only set if child of concrete embedded document
+        """
+        concrete_not_children = [
+            name for name, ed in self._embedded_lookup.items()
+            if not ed.opts.is_child and not ed.opts.abstract
+        ]
+
+        for doc_cls in self._doc_lookup.values():
+            if doc_cls.opts.abstract:
+                continue
+            if doc_cls.opts.is_child:
+                continue
+            for doc in doc_cls.collection.find():
+                doc = remove_cls_field_from_embedded_docs(doc, concrete_not_children)
+                ret = doc_cls.collection.replace_one({"_id": doc["_id"]}, doc)
+                if ret.matched_count != 1:
+                    raise UpdateError(ret)
