@@ -1,3 +1,6 @@
+from .abstract import BaseMarshmallowSchema
+
+
 class MetaTemplate(type):
 
     def __new__(cls, name, bases, nmspc):
@@ -18,6 +21,8 @@ class Template(metaclass=MetaTemplate):
     """
     Base class to represent a template.
     """
+    MA_BASE_SCHEMA_CLS = BaseMarshmallowSchema
+
     def __init__(self, *args, **kwargs):
         raise NotImplementedError('Cannot instantiate a template, '
                                   'use instance.register result instead.')
@@ -32,8 +37,7 @@ class MetaImplementation(MetaTemplate):
         if 'opts' not in nmspc:
             # Inheritance to avoid metaclass conflicts
             return super().__new__(cls, name, bases, nmspc)
-        else:
-            return type.__new__(cls, name, bases, nmspc)
+        return type.__new__(cls, name, bases, nmspc)
 
     def __repr__(cls):
         return "<Implementation class '%s.%s'>" % (cls.__module__, cls.__name__)
@@ -45,13 +49,12 @@ class Implementation(metaclass=MetaImplementation):
     """
     @property
     def opts(self):
-        "An implementation must provide it configuration though this attribute."
+        "An implementation must provide its configuration though this attribute."
         raise NotImplementedError()
 
 
 def get_template(template_or_implementation):
     if issubclass(template_or_implementation, Implementation):
         return template_or_implementation.opts.template
-    else:
-        assert issubclass(template_or_implementation, Template)
-        return template_or_implementation
+    assert issubclass(template_or_implementation, Template)
+    return template_or_implementation
